/*
 * Copyright (c) 2005 Mellanox Technologies.  All rights reserved.
 * Copyright (c) 2005 SilverStorm Technologies.  All rights reserved.
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: PrintIp.c 2897 2010-09-06 15:44:30Z tzachid $
 */


#include <windows.h>
#include <devioctl.h>
#include <stdio.h>
#include <Iphlpapi.h>
#include "iba\ib_types.h"
#include <iba\ib_at_ioctl.h>


// Print all ips that are related to infiniband on this computer
int print_ips()
{
	HANDLE hKernelLib;
	HRESULT hr = S_OK;
	char temp [1000];
	char temp1 [1000];
	IOCTL_IBAT_PORTS_IN ipoib_ports_in;
	IOCTL_IBAT_PORTS_OUT *p_ipoib_ports_out;
	IBAT_PORT_RECORD *ports_records;

	IOCTL_IBAT_IP_ADDRESSES_IN addresses_in;
	IOCTL_IBAT_IP_ADDRESSES_OUT *addresses_out;
	IP_ADDRESS       *ip_addreses;

	BOOL ret;
	int i,j;
	DWORD BytesReturned = 0;

	printf("Adapters that are known to the ipoib modules are:\n\n");

	hKernelLib =
		CreateFileW(
		IBAT_WIN32_NAME,
		GENERIC_READ | GENERIC_WRITE,
		FILE_SHARE_READ | FILE_SHARE_WRITE, // share mode none
		NULL,                               // no security
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL                                // no template
		);

	if (hKernelLib == INVALID_HANDLE_VALUE) {
		hr = HRESULT_FROM_WIN32(GetLastError());
		printf("failed to open the kernel device hr=0x%x\n", hr);
		return 1;
	}
    RtlSecureZeroMemory(&ipoib_ports_in, sizeof (ipoib_ports_in));
    RtlSecureZeroMemory(&addresses_in, sizeof(addresses_in));
	ipoib_ports_in.Version = IBAT_IOCTL_VERSION;

	p_ipoib_ports_out = (IOCTL_IBAT_PORTS_OUT *)temp;

	ret = DeviceIoControl(
		hKernelLib,
		IOCTL_IBAT_PORTS,
		&ipoib_ports_in,
		sizeof(ipoib_ports_in),
		p_ipoib_ports_out,
		sizeof(temp),
		&BytesReturned,
		NULL
		);

	if (ret == 0) {
		hr = HRESULT_FROM_WIN32(GetLastError());
		printf("DeviceIoControl failed for IOCTL_IBAT_PORTS hr=0x%x\n", hr);
		return 1;
	}
	if (p_ipoib_ports_out->Size > sizeof(temp)) {
		printf("Data truncated, please call again with a buffer of %d bytes", p_ipoib_ports_out->Size);
	}

	ports_records = p_ipoib_ports_out->Ports;
	printf("Number of devices %d\n", p_ipoib_ports_out->NumPorts);
	for (i = 0 ; i < p_ipoib_ports_out->NumPorts; i++)
	{
		printf("%d: ca guid = 0x%I64x port guid=0x%I64x\n", i, CL_NTOH64(ports_records[i].CaGuid), CL_NTOH64(ports_records[i].PortGuid));

		// print the ip adresses of this port
		addresses_in.Version = IBAT_IOCTL_VERSION;
		addresses_in.PortGuid = ports_records[i].PortGuid;

		addresses_out = (IOCTL_IBAT_IP_ADDRESSES_OUT *)temp1;

		ret = DeviceIoControl(
			hKernelLib,
			IOCTL_IBAT_IP_ADDRESSES,
			&addresses_in,
			sizeof(addresses_in),
			addresses_out,
			sizeof(temp1),
			&BytesReturned,
			NULL
			);

		if (ret == 0)
		{
			hr = HRESULT_FROM_WIN32(GetLastError());
			printf("DeviceIoControl failed for IOCTL_IBAT_IP_ADDRESSES hr=0x%x\n", hr);
			return 1;
		}
		if (addresses_out->Size > sizeof(temp1) )
		{
			printf("Data truncated, please call again with a buffer of %d bytes", p_ipoib_ports_out->Size);
			return 1;
		}

		printf("   found %d ips:", addresses_out->AddressCount);
		ip_addreses = addresses_out->Address;
		for (j = 0; j < addresses_out->AddressCount; j++)
		{
			printf("    %d.%d.%d.%d   ",
				ip_addreses[j].Address[12],
				ip_addreses[j].Address[13],
				ip_addreses[j].Address[14],
				ip_addreses[j].Address[15]);
		}
		printf("\n");
	}

	return 0;
};

void print_usage(char *argv[])
{
	printf("This program is used to print ip adapters and their addresses or to do arp\n");
	printf("Usage is: %s <print_ips> \n",argv[0]);
	printf("or %s <remoteip> <ip>  (for example %s remoteip 1.2.3.4)\n", argv[0],argv[0]);
}

int remote_ip(char *remote_ip)
{
	HANDLE hKernelLib;
	HRESULT hr = S_OK;
	IPAddr ip;
	char *pIp = (char *)&ip;
	int b1,b2,b3,b4;
	DWORD  ret;
	IOCTL_IBAT_MAC_TO_GID_IN mac;
	IOCTL_IBAT_MAC_TO_GID_OUT gid;
	DWORD BytesReturned = 0;

	ULONG pMacAddr[2], PhyAddrLen ;
	unsigned char *pMac = (unsigned char *)&pMacAddr;
	PhyAddrLen = sizeof(pMacAddr);
    RtlSecureZeroMemory(&mac, sizeof(mac));
	hKernelLib =
		CreateFileW(
		IBAT_WIN32_NAME,
		GENERIC_READ | GENERIC_WRITE,
		FILE_SHARE_READ | FILE_SHARE_WRITE, // share mode none
		NULL,                               // no security
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL                                // no template
		);

	if (hKernelLib == INVALID_HANDLE_VALUE)
	{
		hr = HRESULT_FROM_WIN32(GetLastError());
		printf("failed to open the kernel device hr=0x%x\n", hr);
		return 1;
	}

	sscanf(remote_ip, "%d.%d.%d.%d", &b1, &b2, &b3, &b4);
	printf("Calling arp for addresses %d.%d.%d.%d\n", b1, b2, b3, b4);

	pIp[0] = (char)b1;
	pIp[1] = (char)b2;
	pIp[2] = (char)b3;
	pIp[3] = (char)b4;

	ret = SendARP(ip ,0 ,pMacAddr, &PhyAddrLen );
	if (ret != NO_ERROR)
	{
		printf("Error in SendARP");
		return 1;
	}

	printf("Mac of the remote addresses is %x-%x-%x-%x-%x-%x\n",
		pMac[0], pMac[1], pMac[2], pMac[3], pMac[4], pMac[5] );

	// query for the gid
	memcpy(mac.DestMac, pMac, 6);

	ret = DeviceIoControl(
		hKernelLib,
		IOCTL_IBAT_MAC_TO_GID,
		&mac,
		sizeof(mac),
		&gid,
		sizeof(gid),
		&BytesReturned,
		NULL );

	if (ret == 0)
	{
		hr = HRESULT_FROM_WIN32(GetLastError());
		printf("DeviceIoControl failed for IOCTL_IBAT_IP_ADDRESSES hr=0x%x\n", hr);
	}

	printf("lid of remote ip is = 0x%I64x : 0x%I64x\n", CL_NTOH64(gid.DestGid.unicast.prefix), CL_NTOH64(gid.DestGid.unicast.interface_id));

	return 0;
}


int __cdecl main(int argc, char *argv[])
{
	if (argc < 2) {
		print_usage(argv);
		return 1;
	}
	if (!strcmp(argv[1], "print_ips")) {
		return print_ips();
	}
	if (!strcmp(argv[1], "remoteip")) {
		return remote_ip(argv[2]);
	}
	print_usage(argv);
	return 1;
}
